#!/usr/bin/env python2
"""
Helpers and definitions for tests for the mod(erator) module.  Many of these
should be applicable for any ISFDB tests, and so are likely to be moved to
a more generic location (assuming these tests become part of the project).
"""

# from collections import namedtuple # Was added in Python 2.6 :-(
from datetime import datetime
import os
import re
import subprocess
import unittest
from urlparse import urlparse # IIRC this changes in Python 3?

# from bs4 import BeautifulSoup # Works on Py2.7, but not 2.5
from BeautifulSoup import BeautifulSoup
import MySQLdb


# FieldDetails = namedtuple('FieldDetails', 'current, proposed, warning')

class FieldDetails(object):
        """Simplistic object to cover namedtuple not existing in Python 2.5"""
        def __init__(self, current, proposed, warning):
                self.current = current
                self.proposed = proposed
                self.warning = warning

        def __eq__(self, other):
                # Grr. this isn't picked up when doing comparisons like
                # self.assertEqual({"x": FieldDetails("a", "b", "c"),
                #                  {"x": FieldDetails("a", "b", "c")
                # (I don't think think assertDictEqual() would help, and
                # it's Python 2.7+ anyway)
                return self.current == other.current and \
                        self.proposed == other.proposed and \
                        self.warning == other.warning

        def __cmp__(self, other): # Not sure if this is needed, but just in case...
                return self.__eq__(other)

        def __ne__(self, other): # Not sure if this is needed, but just in case...
                return not self.__eq__(other)

        def __repr__(self):
                return 'FieldDetails(current=%s, proposed=%s, warning=%s' % \
                        (self.current, self.proposed, self.warning)



class UnexpectedSuccessError(Exception):
        pass

# TODO: replace the code that uses this with the tempfile stdlib module
TEMP_DIR = '/tmp' # Might be something different on Windows?


DEFAULT_PYTHON_EXECUTABLE = 'python2'

THIS_DIR = os.path.dirname(__file__)


def make_cookie(username, user_id):
        return 'isfdbUserID=%s; isfdbUserName=%s; isfdbToken=' % (user_id, username)

def split_headers_and_html_from_cgi_response(content):
        headers = []
        hdr, remainder = content.split('\n', 1)
        while len(hdr) > 0:
                headers.append(hdr)
                hdr, remainder = remainder.split('\n', 1)
        # TODO (maybe): convert headers into a dict?
        return headers, remainder

def check_output(arg_list, env):
        """Because Python 2.5 doesn't have subprocess.check_output() :-("""
        my_proc = subprocess.Popen(arg_list, env=env, stdout=subprocess.PIPE)
        content = my_proc.stdout.read()
        # print('Returncode is %s' % my_proc.returncode) # prints None
        wait_thing = my_proc.wait()
        # print('wait_thing is %s' % wait_thing)
        if wait_thing != 0:
                # Great - Python 2.5.1 doesn't have the content argument :-(
                ex = subprocess.CalledProcessError(wait_thing, # returncode
                                                    'Blah') # cmd?
                ex.output = content
                raise ex
                #raise subprocess.CalledProcessError(wait_thing, # returncode
                #                                    'Blah', # cmd?
                #                                    content) # output
        return content

def capture_html_from_cgi_script_as_soup(arg_list, run_as='nobody', env=None,
                                         expect_error=False):
        if os.path.basename(arg_list[0]) not in ('python', 'python2', 'python3'):
                arg_list.insert(0, DEFAULT_PYTHON_EXECUTABLE)
        env_vars = {'PYTHONPATH': '%s:%s' % (THIS_DIR, os.path.join(THIS_DIR,
                                                                    '..',
                                                                    'local'))}
        if env:
                env_vars.update(env)
        if expect_error:
                try:
                        content = check_output(arg_list,
                                               env=env_vars)
                        raise UnexpectedSuccessError('Expected failure running %s, but it succeeded?!?' %
                                                     (arg_list))
                except subprocess.CalledProcessError, err:
                        content = err.output
        else:
                content = check_output(arg_list,
                                       env=env_vars)
        _, html_content = split_headers_and_html_from_cgi_response(content)
        # soup = BeautifulSoup(html_content, 'lxml') # bs4 only?
        soup = BeautifulSoup(html_content)
        return soup

class TableCell(object):
        """
        Class representing table cell textual content and classes.  Designed
        to make it easy to test against both properties when you care about
        both of them, or only text when that's all that's relevant.

        >>> TableCell('foo', ['c1']) == TableCell('foo', ['c1', 'c2'])
        True
        >>> TableCell('foo', ['c1']) == TableCell('foo', ['c2'])
        False
        >>> TableCell('foo', ['c1']) == 'foo'
        True
        """

        def __init__(self, text, css_classes=None):
                self.text = text
                if css_classes:
                        # older BS returns a string/unicode
                        if isinstance(css_classes, (str, unicode)):
                                css_classes = [z for z in re.split('\s+', css_classes)
                                               if z]
                        self.css_classes = set(css_classes)
                else:
                        self.css_classes = set()

        def __eq__(self, other):
                if isinstance(other, TableCell):
                        if not self.css_classes or not other.css_classes:
                                return self.text.strip() == other.text.strip()
                        return self.text.strip() == other.text.strip() and \
                                (self.css_classes.issuperset(other.css_classes) or
                                 self.css_classes.issubset(other.css_classes))
                else: # Assume other is a string or Unicode
                        return self.text == other

        def __repr__(self):
                return 'TableCell: text=%s; classes=%s' % (self.text, self.css_classes)


def sanitize_cell_text(cell, normalize_spaces=True, remove_ids=False):
        if normalize_spaces:
                # It seems bs4 converts &nbsp; to space, but older
                # versions don't?
                # Also, older versions don't take into account <br> tags,
                # so "Foo<br />Bar" becomes "FooBar" rather than "Foo\nBar"
                for elem in cell.findAll('br'):
                        elem.replaceWith('&nbsp;') # replacing with \n is no good as that gets eaten
                txt = cell.text.replace('&nbsp;', ' ')
        else:
                txt = cell.text
        if remove_ids:
                txt = re.sub('\[.*\]', '',  txt)
        ret = txt.strip()
        if ret == 'Proposed Changes': # e.g. pv_update.cgi
                ret = 'Proposed' # e.g. av_update.cgi
        return ret


def parse_table(soupified_table, remove_ids_from_column_names=True):
        """
        Given a BS table object, return a list of dicts for each data row of the
        table.
        """
        column_headings = []
        ret = []
        for i, raw_row in enumerate(soupified_table.findAll('tr')):
                # We can't rely on the use of <th> over <td> (or the presence
                # of <thead> vs <tbody>, so just assume the first row is headings
                if i == 0:
                        # This will screw up if <th> and <td> are intermixed
                        cells = raw_row.findAll('th') + raw_row.findAll('td')
                        cell_text = [sanitize_cell_text(z, remove_ids=remove_ids_from_column_names)
                                     for z in cells]
                        column_headings = cell_text
                else:
                        raw_cells = raw_row.findAll('td')
                        cell = [TableCell(sanitize_cell_text(z), z.get('class')) for z in raw_cells]
                        row = dict(zip(column_headings, cell))
                        ret.append(row)
        return ret

def parse_proposed_table(soupified_table):
        """
        Given a BS table object with 'Column', 'Proposed Values' and 'Warnings'
        columns, return two dicts, one mapping the column names to proposed values
        and the other mapping column names to warnings.  For the latter, only
        keys/"columns" which contain a value are included.
        """
        row_dicts = parse_table(soupified_table)
        proposed_dict = dict([(d['Column'].text, d['Proposed Values']) for d in row_dicts])
        raw_warnings = [(d['Column'].text, d['Warnings']) for d in row_dicts]
        warning_dict = dict([z for z in raw_warnings if z[1].text])
        return proposed_dict, warning_dict


def get_heading_ids(column_headings):
        """
        Given an iterable of column headings containing IDs, return a dict mapping
        generic name to a tuple of (numeric ID, original column name)

        e.g.

        >>> get_heading_ids(['KeepId [1155736]', 'DropId [2015329]'])
        {'KeepId': (1155736, 'KeepId [1155736]'), 'DropId': (2015329, 'DropId [2015329]')}
        """
        ret = {}
        for heading in column_headings:
                # Q: (How) should this handle columns without IDs?
                bits = re.match('(.*) \[(\d+)\]', heading)
                ret[bits.group(1)] = (int(bits.group(2)), heading)
        return ret

def parse_merge_table(soupified_table):
        """
        Given a BS table object with 'Column', 'KeepId [nnnn]' and 'DropId [mmmm]'
        columns, return a tuple of (int, dict, int, dict) comprising
        * the ID of the title to keep
        * dict mapping the column names to keep values
        * and similarly for the drop ID and values
        """
        row_dicts = parse_table(soupified_table, remove_ids_from_column_names=False)
        heading_details = get_heading_ids([z for z in row_dicts[0].keys()
                                           if z != 'Column'])

        keep_heading = heading_details['KeepId'][1]
        keep_dict = dict([(d['Column'].text, d[keep_heading]) for d in row_dicts])
        drop_heading = heading_details['DropId'][1]
        drop_dict = dict([(d['Column'].text, d[drop_heading]) for d in row_dicts])

        return (heading_details['KeepId'][0], keep_dict,
                heading_details['DropId'][0], drop_dict)




def parse_change_table(soupified_table):
        """
        Given a BS table object from a moderator approval page, returns two
        dicts:

        * changed: maps property/row name to
                   FieldDetails(current-value, proposed-value, warning)
        * unchanged: maps property/row name to current-value
        """
        column_headings = []
        changed = {}
        unchanged = {}
        # TODO: support <th> and/or <thead> if they exist (currently these aren't
        # in the HTML response, but ideally they should be).

        for i, raw_row in enumerate(soupified_table.findAll('tr')):
                cells = raw_row.findAll('td')
                if i == 0:
                        column_headings = [sanitize_cell_text(z, remove_ids=True) for z in cells]
                else:
                        cell_text = [sanitize_cell_text(z) for z in cells]
                        row = dict(zip(column_headings, cell_text))
                        prop = row['Column']
                        if row['Proposed'] in ('', '-'):
                                unchanged[prop] = row['Current']
                        else:
                                changed[prop] = FieldDetails(row['Current'],
                                                             row['Proposed'],
                                                             row['Warnings'])

        return changed, unchanged

def verify_button_link(soupified_button, path, query):
        # "button" is a slight misnomer - this is for  <a> links that are styled
        # (a bit) like buttons.
        url_bits = urlparse(soupified_button['href'])
        return (path == url_bits.path) and (str(query) == str(url_bits.query))

def load_xml_file(filename):
        if not os.path.exists(filename):
                filename = os.path.join(os.path.dirname(__file__), filename)
        xml_inputstream = open(filename)
        xml_content = xml_inputstream.read()
        xml_inputstream.close()
        return xml_content


def Date_or_None(s):
        return s

def IsfdbConvSetup():
        import MySQLdb.converters
        IsfdbConv = MySQLdb.converters.conversions
        IsfdbConv[10] = Date_or_None
        return(IsfdbConv)

def setup_database(db_host, db_user, db_password, database):
        """
        Return a connection object for specified MySQL/MariaDB database.
        """
        db = MySQLdb.connect(db_host, db_user, db_password,
                             conv=IsfdbConvSetup())
        db.select_db(database)
        return db



class ISFDBTestCase(unittest.TestCase):

        @property
        def moderator_id(self):
                moderator_id = self.run_sql_statement("""SELECT user_id
                FROM mw_user WHERE user_name = 'TestModerator';""")
                return moderator_id

        @property
        def editor_id(self):
                editor_id = self.run_sql_statement("""SELECT user_id
                FROM mw_user WHERE user_name = 'TestEditor';""")
                return editor_id

        @property
        def user_cookies(self):
                moderator_id = self.moderator_id # avoid doing 2 DB lookups below
                editor_id = self.editor_id # avoid doing 2 DB lookups below

                # The duplicate keys/names are just a convenience for historical reasons
                return {
                        'nobody': '',
                        'editor': make_cookie('TestEditor', editor_id),
                        'TestEditor': make_cookie('TestEditor', editor_id),
                        'moderator': make_cookie('TestModerator', moderator_id),
                        'TestModerator': make_cookie('TestModerator', moderator_id)
                        # TODO (maybe): bureaucrat also?
                }

        def run_script_as_user(self, arg_list, run_as, expect_error=False):
                env = {'HTTP_COOKIE': self.user_cookies[run_as]}
                try:
                        env['PYTHONPATH'] = '%s:%s' % (self.localdefs_path,
                                                       os.path.join(THIS_DIR,
                                                                    '..',
                                                                    'local'))
                except AttributeError:
                        pass # use the default PYTHONPATH set in the capture...() function
                return capture_html_from_cgi_script_as_soup([str(z) for z in arg_list],
                                                            env=env,
                                                            expect_error=expect_error)

        def run_sql_statement(self, query, commit=True):
                # This requires the class to be initialized with a ISFDB_Database_Base
                # mixin to provide the setup_database() method.
                db = self.setup_database()
                db.query(query)
                if query.lower().strip().startswith('insert'):
                        retval = db.insert_id()
                elif query.lower().strip().startswith('select'):
                        result = db.store_result()
                        record = result.fetch_row()
                        retval = record[0][0] # Assumption: we only want a single ID back
                else:
                        retval = None
                if commit:
                        db.commit()
                db.close()
                return retval


class ISFDBSubmissionsTestCase(ISFDBTestCase):
        """
        Class for testing CGI scripts based on the submissions table e.g.
        those in the mod/ directory called *_(new|update|merge).py

        Classes derived from this should:
        * define xml_filename and sub_type class variables
        * pull in an appropriate ISFDB_yyyy_mm_dd_Mixin

        Q: Should this include the ModeratorActionsFailAsNonModeratorMixin
        by default?  I can't imagine there are any of these pages that aren't
        moderator-only?
        """

        def setUp(self):
                xml_content = load_xml_file(self.xml_filename)
                now = datetime.now().isoformat()[:19].replace('T', ' ')
                sql = """INSERT submissions(sub_state, sub_type, sub_data,
                sub_time, sub_submitter)
                VALUES ('N', %d, '%s', '%s', %d);""" % (self.sub_type,
                                                        xml_content, now,
                                                        self.moderator_id)
                self.sub_id = self.run_sql_statement(sql)

        def tearDown(self):
                sql = 'DELETE FROM submissions WHERE sub_id = %d;' % (self.sub_id)
                self.run_sql_statement(sql)

        def verify_hold_and_approval_buttons(self, soup, approval_link,
                                         hold_link='/cgi-bin/mod/hold.cgi'):
                # hold_button = soup.select_one('a.hold') # bs4
                hold_button = soup.find('a', {'class': 'hold'})
                self.assertTrue(verify_button_link(hold_button,
                                                   hold_link,
                                                   self.sub_id))
                # approval_button = soup.select_one('a.approval') # bs4
                approval_button = soup.find('a', {'class': 'approval'})
                self.assertTrue(verify_button_link(approval_button,
                                                   approval_link,
                                                   self.sub_id))




class ISFDB_Database_Base(object):
        """
        Do not use this class directly in tests - instead use an
        ISFDB_YYYY_MM_DD_Mixin that is derived from this class.
        """
        def setup_database(self):
                return setup_database(self.db_host, self.db_user, self.db_password,
                                      self.database)

        def write_localdefs_file(self):
                temp_dir = os.path.join(TEMP_DIR, '%s_testing' % self.database)
                try:
                        os.makedirs(temp_dir) # exist_ok arg only as of Python 3.2
                except: # OSError in Python2, FileExistsError in Python3
                        pass
                filename = os.path.join(temp_dir, 'localdefs.py')
                outputstream = open(filename, 'w')

                txt = '''# This is an auto-generated file created by %s
HTMLLOC = "localhost"
HTFAKE = "/localhost/cgi-bin"
DBASEHOST = "%s"
HTMLHOST = "localhost"
COOKIEHOST = "localhost"
WIKILOC = "localhost/wiki"


# For most tests, it's the next three values that are the only ones that will
# be used
USERNAME = "%s"
PASSWORD = "%s"
DBASE = "%s"

UNICODE = "iso-8859-1"
DO_ANALYTICS = 0
''' % (__file__, self.db_host, self.db_user, self.db_password, self.database)

                outputstream.write(txt)
                outputstream.close()
                return temp_dir # This will be useful for setting PYTHONPATH

class ISFDB_2019_10_19_Mixin(ISFDB_Database_Base):
        db_host = os.environ.get('ISFDB_TEST_DATABASE_HOST', 'localhost')
        db_user = 'readwrite'
        db_password = 'readwrite'
        database = 'isfdb_2019_10_19'

        @property
        def localdefs_path(self):
                return self.write_localdefs_file()

# define more mixins to use a different database


class ModeratorActionsFailAsNonModeratorMixin(object):
        def test_action_fails_as_nobody(self):
                # TODO: ideally this page should respond with HTTP 401 Unauthorized
                soup = self.run_script_as_user([self.script, str(self.sub_id)],
                                               run_as='nobody')
                # main2_heading = soup.select_one('#main2 h2') # bs4
                main2 = soup.find('div', {'id': 'main2'})
                main2_heading = main2.find('h2')
                self.assertEqual('Login required to moderate changes',
                                 main2_heading.text.strip())

        def test_action_fails_as_editor(self):
                # TODO: ideally this page should respond with HTTP 403 Forbidden
                soup = self.run_script_as_user([self.script, str(self.sub_id)],
                                               run_as='editor')
                # main2_heading = soup.select_one('#main2 h2') # bs4
                main2 = soup.find('div', {'id': 'main2'})
                main2_heading = main2.find('h2')
                self.assertEqual('Moderator privileges are required for this option',
                                 main2_heading.text.strip())

        def test_action_with_unknown_id_as_nobody(self):
                # TODO: ideally this page should respond with HTTP 401 Unauthorized
                soup = self.run_script_as_user([self.script, str(self.sub_id)],
                                               run_as='nobody')
                # main2_heading = soup.select_one('#main2 h2') # bs4
                main2 = soup.find('div', {'id': 'main2'})
                main2_heading = main2.find('h2')
                self.assertEqual('Login required to moderate changes',
                                 main2_heading.text.strip())

        def test_action_with_unknown_id_as_moderator(self):
                # TODO: currently this responds with status 200 but a page
                # with a Python stack trace.  Ideally it should be status 404
                # and a nicer error message (unless debug mode enabled?)
                soup = self.run_script_as_user([self.script, 1],
                                               run_as='moderator',
                                               expect_error=True)
                paras = soup.findAll('p')
                self.assertTrue(any(['A problem occurred in a Python script' in z.text
                                     for z in paras]))


if __name__ == '__main__':
        import doctest
        doctest.testmod()
